import sys
from os import path
import os
import re
import shutil
import json
from langchain.output_parsers import json as json_parser
sys.path.append("../")

from langchain.chains.summarize import load_summarize_chain
from langchain.chains.question_answering import load_qa_chain
from langchain.schema.language_model import BaseLanguageModel
from langchain.prompts import PromptTemplate
from langchain.chains import RetrievalQA
from typing import List,Dict,Optional
from langchain.schema import Document
from langchain.vectorstores.base import VectorStoreRetriever
from langchain.vectorstores.utils import maximal_marginal_relevance
from langchain.callbacks.manager import (
    Callbacks,
)
from langchain.embeddings.base import Embeddings
from langchain.embeddings.huggingface import (
    HuggingFaceEmbeddings,
    HuggingFaceInstructEmbeddings,  #pip install InstructorEmbedding
)
from llm.chatglm import ChatGLMSerLLM
from llm.ernie_with_sdk import ChatERNIESerLLM
from vector.pgsqldocstore import PgSqlDocstore
from langchain.chains.router.llm_router import LLMRouterChain

from langchain.vectorstores import FAISS
from langchain.docstore.in_memory import InMemoryDocstore
import faiss
from loader import load
from scenarios.spdsvb.similarity import VectorStore_FAISS
from .prompts import (
    QA_PROMPT,
    REFINE_QA_PROMPT,
    SUMMARISE_PROMPT,
    REFINE_SUMMARISE_PROMPT,
    EXTRACTION_PROMPT,
    REFINE_EXTRACTION_PROMPT,
    ROUTER_PROMPT,
    GLM_CHAT_QUESTION_PROMPT,
    GLM_CHAT_COMBINE_PROMPT,
    ERNIE_GLM_CHAT_QUESTION_PROMPT,
    ERNIE_GLM_CHAT_COMBINE_PROMPT,
    ERNIE_EXTRACTION_PROMPT,
    ERNIE_REFINE_EXTRACTION_PROMPT,
    MAP_REDUCE_SUMMARISE_PROMPT,
    GLM_MAP_EXTRACTION_PROMPT,
    GLM_MAP_EXTRACTION_COMBINE_PROMPT,
    FOREACH_REFINE_EXTRACTION_PROMPT,
    FOREACH_EXTRACTION_PROMPT,
    FOREACH_MAP_EXTRACTION_PROMPT,
    FOREACH_MAP_Q_EXTRACTION_PROMPT
)

def del_vectorstore_path(index_name:str,vectorstorepath:str="./vectorstore"):
    if index_name:
        filepath = path.join(vectorstorepath,index_name)
        if path.exists(filepath):
            shutil.rmtree(filepath)

class DocumentQA():
    r"""
        基于文档问答的封装类
    """ 
    def __init__(self,llm: BaseLanguageModel):
        self.llm = llm
    # 文档摘要提取
    def summarize_document(
            self,
            filepaths:[str] = [],
            load_kwargs: Optional[dict] = None,
            documents:List[Document] = [],
            chain_type:str = "map_reduce",
            chain_type_kwargs: Optional[dict] = {},
            **kwargs) -> str:
        r"""
            对文档进行总结比较好：参数说明
            filepath：需要总结的文件路径可为空
            load_kwargs：文档载入参数，查阅load.load方法
            documents：需要总结的文档
            chain_type：链类型，总结方法不同，"stuff","map_reduce","refine"
            chain_type_kwargs:{
                question_prompt：提问prompt
                refine_prompt：refine提问prompt
                document_variable_name：文档关键字
                ...
            }
            kwargs： {
                callbacks
            }
        """ 
        default_chain_type_kwargs = dict
        if chain_type == "map_reduce":
            default_chain_type_kwargs = {"map_prompt":MAP_REDUCE_SUMMARISE_PROMPT,"combine_prompt":MAP_REDUCE_SUMMARISE_PROMPT}
        else:
            chain_type = "refine"
            default_chain_type_kwargs = {"question_prompt":SUMMARISE_PROMPT,"refine_prompt":REFINE_SUMMARISE_PROMPT,"document_variable_name":"text"}
        default_chain_type_kwargs.update(**chain_type_kwargs)
        effect_documents = []
        if len(filepaths) <= 0 and len(documents) <= 0:
            raise ValueError(f"file and documents cannot be empty at the same time")
        if len(filepaths) > 0:
            load_documents= load.loads(filepaths,**load_kwargs)
            effect_documents.extend(load_documents)
        if len(documents) > 0:
            effect_documents.extend(documents)
        # 拼接段文本，不改变原始排版
        effect_documents = load.append(effect_documents,sentence_size=load_kwargs["sentence_size"])
        chain = load_summarize_chain(self.llm, chain_type=chain_type,**default_chain_type_kwargs)
        try:
            result= chain.run(input_documents=effect_documents,**kwargs)
            # print(result)
            return result
        except Exception as e:
            print(e)
            raise ValueError(f'文档总结报错：{e}')
        
    # 文档提问
    def qa_from_document(
            self,
            query:str,
            retriever:VectorStoreRetriever,
            chain_type:str = "map_reduce",
            chain_type_kwargs: Optional[dict] = {},
            chain_kwargs: Optional[dict] = {},
            **kwargs) -> str:
        r"""
            对文档进行提问比较好：参数说明
            query：问题
            retriever：向量库
            chain_type：链类型，总结方法不同，"stuff","map_reduce","refine"
            chain_type_kwargs:{  根据chain类型提供指定参数
                question_prompt：提问prompt
                refine_prompt：refine提问prompt
                document_variable_name：文档关键字
                verbose：debug日志
                ...
            }
            chain_kwargs:{}
            kwargs：{
                callbacks
            }
        """ 
        #设置默认链参数
        default_chain_type_kwargs = dict
        if chain_type == "map_reduce":
            question_prompt=GLM_CHAT_QUESTION_PROMPT
            combine_prompt=GLM_CHAT_COMBINE_PROMPT
            if isinstance(self.llm,ChatERNIESerLLM):
                question_prompt=ERNIE_GLM_CHAT_QUESTION_PROMPT
                combine_prompt=ERNIE_GLM_CHAT_COMBINE_PROMPT
            default_chain_type_kwargs = {"question_prompt":question_prompt,"combine_prompt":combine_prompt}
        else:
            chain_type = "refine"
            default_chain_type_kwargs = {"question_prompt":QA_PROMPT,"refine_prompt":REFINE_QA_PROMPT,"document_variable_name":"context"}
        default_chain_type_kwargs.update(**chain_type_kwargs)
        qa = RetrievalQA.from_chain_type(llm=self.llm, chain_type=chain_type, 
                                 chain_type_kwargs=default_chain_type_kwargs,retriever=retriever,**chain_kwargs)
        try:
            result = qa.run(query=query,**kwargs)
            # print(result)
            return result
        except Exception as e:
            print(e)
            raise ValueError(f'文档问答报错：{e}')
        
    # 文档提取
    # 注意：
    #       若使用GLM，建议调用extract_from_document_foreach，同时chain_type=refine
    #       若使用ERNIE，建议调用extract_from_document，同时chain_type=refine
    def extract_from_document(
            self,
            query:str,
            filepaths:[str] = [],
            load_kwargs: Optional[dict] = None,
            documents:List[Document] = [],
            chain_type:str = "refine",
            chain_type_kwargs: Optional[dict] = {},
            **kwargs) -> json:
        r"""
            从文档中提取关键信息：参数说明
            query：需要提取的知识点
            filepaths：需要总结的文件路径可为空
            load_kwargs：文档载入参数，查阅load.load方法
            documents：需要总结的文档
            chain_type：链类型，总结方法不同，"stuff","map_reduce","refine"
            chain_type_kwargs:{
                verbose：debug日志
                question_prompt：提问prompt
                refine_prompt：refine提问prompt
                document_variable_name：文档关键字
            }
            kwargs：{
                callbacks：run回调
            }
        """ 
        #设置默认链参数
        default_chain_type_kwargs = dict
        if chain_type == "map_reduce":
            question_prompt=GLM_MAP_EXTRACTION_PROMPT
            combine_prompt=GLM_MAP_EXTRACTION_COMBINE_PROMPT
            if isinstance(self.llm,ChatERNIESerLLM):
                # todo
                raise ValueError(f"Ernie's performance is not good when 'chain_type=map-reduce'")
            default_chain_type_kwargs = {"question_prompt":question_prompt,"combine_prompt":combine_prompt}
        else:
            chain_type = "refine"
            question_prompt=EXTRACTION_PROMPT
            refine_prompt=REFINE_EXTRACTION_PROMPT
            if isinstance(self.llm,ChatERNIESerLLM):
                question_prompt=ERNIE_EXTRACTION_PROMPT
                refine_prompt=ERNIE_REFINE_EXTRACTION_PROMPT
            default_chain_type_kwargs = {"question_prompt":question_prompt,"refine_prompt":refine_prompt,"document_variable_name":"context"}
        default_chain_type_kwargs.update(**chain_type_kwargs)
        effect_documents = []
        if len(filepaths) <= 0 and len(documents) <= 0:
            raise ValueError(f"file and documents cannot be empty at the same time")
        if len(filepaths) > 0:
            load_documents= load.loads(filepaths,**load_kwargs)
            effect_documents.extend(load_documents)
        if len(documents) > 0:
            effect_documents.extend(documents)
        effect_documents = load.append(documents=effect_documents,sentence_size=load_kwargs["sentence_size"])
        chain = load_qa_chain(llm=self.llm,chain_type=chain_type,**default_chain_type_kwargs)
        
        try:
            result = chain.run(input_documents=effect_documents, question=query,**kwargs)
            # print(result)
            if isinstance(self.llm,ChatERNIESerLLM):
                try:
                    return json_parser.parse_json_markdown(json_string=result)
                except Exception as e:
                    print(result)
                    print("序列化报错",e)
                    return json()
            else:
                return result
        except Exception as e:
            print(e)
            raise ValueError(f'文档提取报错：{e}')
        

    def extract_from_document_foreach(
            self,
            query:str,
            pattern:str =  r'[、]',
            filepaths:[str] = [],
            load_kwargs: Optional[dict] = None,
            documents:List[Document] = [],
            chain_type:str = "refine",
            chain_type_kwargs: Optional[dict] = {},
            **kwargs) -> json:
        r"""
            从文档中提取关键信息：参数说明
            query：需要提取的知识点
            pattern：知识点分割符号
            filepaths：需要总结的文件路径可为空
            load_kwargs：文档载入参数，查阅load.load方法
            documents：需要总结的文档
            chain_type：链类型，总结方法不同，"stuff","map_reduce","refine"
            chain_type_kwargs:{
                verbose：debug日志
                question_prompt：提问prompt
                refine_prompt：refine提问prompt
                document_variable_name：文档关键字
            }
            kwargs：{
                callbacks：run回调
            }
        """
        #设置默认链参数
        default_chain_type_kwargs = dict
        if isinstance(self.llm,ChatERNIESerLLM):
            raise ValueError(f"Ernie did not perform well when calling the 'extract_from_document_foreach' function")
        if chain_type == "map_reduce":
            question_prompt=FOREACH_MAP_Q_EXTRACTION_PROMPT         #使用提问的prompt提取关键信息原始文本
            combine_prompt=FOREACH_MAP_EXTRACTION_PROMPT            #从历史提取信息中回答问题
            default_chain_type_kwargs = {"question_prompt":question_prompt,"combine_prompt":combine_prompt}
        else:
            chain_type = "refine"
            default_chain_type_kwargs = {"question_prompt":FOREACH_EXTRACTION_PROMPT,"refine_prompt":FOREACH_REFINE_EXTRACTION_PROMPT,"document_variable_name":"context"}
        default_chain_type_kwargs.update(**chain_type_kwargs)
        effect_documents = []
        if len(filepaths) <= 0 and len(documents) <= 0:
            raise ValueError(f"file and documents cannot be empty at the same time")
        if len(filepaths) > 0:
            load_documents= load.loads(filepaths,**load_kwargs)
            effect_documents.extend(load_documents)
        if len(documents) > 0:
            effect_documents.extend(documents)
        effect_documents = load.append(documents=effect_documents,sentence_size=load_kwargs["sentence_size"])
        chain = load_qa_chain(llm=self.llm,chain_type=chain_type,**default_chain_type_kwargs)
        words_list = re.split(pattern, query)
        results = {}
        for word in words_list:
            try:
                result = chain.run(input_documents=effect_documents, question=word,**kwargs)
                result = result.replace("\n", "")
                result_s = re.split(r'[:：]',result)
                if len(result_s)>=2:
                    results[result_s[0]] = ":".join(result_s[1:])
            except Exception as e:
                print(f'提取信息：{word}报错：\n',e)
        # print(results)
        return json.dumps(results,ensure_ascii=False)
    
    def route_chain(self,
            query:str,
            filepaths:[str] = [],
            load_kwargs: Optional[dict] = {},
            documents:List[Document] = [],
            embeddings_model:str = None,
            chain_type:str = "refine",
            chain_type_kwargs: Optional[dict] ={},
            chain_kwargs: Optional[dict] ={},
            **kwargs) -> any:
        r"""
            判断问题属于那种类型，自动使用相应问答模型进行回答，默认使用qa模型
            query：问题
            filepaths：需要总结的文件路径可为空
            load_kwargs：文档载入参数，查阅load.load方法
            documents：需要总结的文档
            embeddings_model：文本向量模型
            chain_type:提问链类型
            chain_type_kwargs：类型参数
            chain_kwargs: qa类型参数
            kwargs：  run args
        """
        
        if query is None or len(query) == 0:
            #默认总结文档
            return self.summarize_document(filepaths=filepaths,documents=documents,load_kwargs=load_kwargs,chain_type=chain_type,chain_type_kwargs=chain_type_kwargs,**kwargs)
        
        router_chain = LLMRouterChain.from_llm(self.llm, ROUTER_PROMPT)
        result = router_chain.route(query)
        if result.destination == "expertor":
            _faiss = VectorStore_FAISS(embedding_model_name=embeddings_model,show_number=5,index_name="chatid")
            if len(filepaths) > 0:
                _faiss._add_documents_from_dir(filepaths=filepaths,load_kwargs={"sentence_size":1024})
            return self.qa_from_document(query=query,retriever=_faiss.as_retriever(),chain_type=chain_type,chain_type_kwargs=chain_type_kwargs,chain_kwargs=chain_kwargs,**kwargs)
        elif result.destination == "extractor":
            return self.extract_from_document(query=query,documents=documents,filepaths=filepaths,load_kwargs=load_kwargs,chain_type=chain_type,chain_type_kwargs=chain_type_kwargs,**kwargs)
        else:
            return self.summarize_document(filepaths=filepaths,documents=documents,load_kwargs=load_kwargs,chain_type=chain_type,chain_type_kwargs=chain_type_kwargs,**kwargs)
        
        