import re
import time

from pydantic import BaseModel
from langchain.prompts import StringPromptTemplate,PromptTemplate
from langchain import LLMChain
from qa.question import QuestionRDF

from similarity import VectorStore_FAISS

prompt_expert_template = """你是浦发硅谷银行网银系统的专家，请帮助解答用户在使用过程中遇到的问题。
{question}
"""

prompt_history_template = """{history}
上面是之前的对话，你可以继续回答用户的问题。
{question}
"""

prompt_enhancement_template = """{similarity}
请结合上述内容回答以下问题，不要提无关内容：
{question}
"""

prompt_enhancement_history_template = """{history}
上面是之前的对话，下面是可参考的内容。
{similarity}
请结合上述内容回答以下问题，不要提无关内容：
{question}
"""

class Chatbot:
    def __init__(self, model, vectorstore_faiss:VectorStore_FAISS,tokenizer=None, base_model=None, base_tokenizer=None, source_prefix=None,re_history=None):
        self.model = model
        self.tokenizer = tokenizer
        self.base_model = base_model
        self.base_tokenizer = base_tokenizer
        self.source_prefix = source_prefix
        self.re_history = re_history
        self.vectorstore_faiss = vectorstore_faiss

    def _build_history(self, history=None):  
        if history is None:
            return None
        prompt=""
        for i, (old_query, response) in enumerate(history):
            prompt += "问：{}\n答：{}\n\n".format(old_query, response)
        return prompt        

    def _build_prompt(self, query, history=None):
        if history is None:
            history = []
        prompt = ""
        for i, (old_query, response) in enumerate(history):
            prompt += "[Round {}]\n\n问：{}\n\n答：{}\n\n".format(i + 1, old_query, response)
        prompt += "[Round {}]\n\n问：{}\n\n答：".format(len(history) + 1, query)
        return prompt

    def rdf_question(self, question, history):
        rdf = QuestionRDF(self.base_model)
        return rdf.generate(history, question)

    def chat_with_llm(self, input, isExpert=False, isEnhancement=False, isBase=False, dialog=[], temp=0.8):        

        run_llm = self.model if isExpert else self.base_model
        history = self._build_history(self.re_history(dialog)) if self.re_history is not None else self._build_history(dialog)

        similarity=None

        # if history is not None and len(history) > 0:
        #     question = self.rdf_question(input, history)
        # else:
        #     question = input
        
        # print("问：", question)
        question = input

        if isExpert: # 专家指令，不带历史
            history, similarity = None, None
            prompt=PromptTemplate.from_template(prompt_expert_template)
        elif isEnhancement: # 知识增强，不带历史，补充相似度
            history, similarity = None, self.vectorstore_faiss._join_document(self.vectorstore_faiss.get_text_similarity_with_score(input))
            if similarity and similarity.strip(): # 有补充知识
                print("相似度：", similarity)
                prompt=PromptTemplate.from_template(prompt_enhancement_template) if history is None else PromptTemplate.from_template(prompt_enhancement_history_template)
            else: # 无补充知识，退化为专家指令
                run_llm = self.model
                prompt=PromptTemplate.from_template(prompt_expert_template)
                # prompt=PromptTemplate.from_template(prompt_history_template) if history is not None else PromptTemplate.from_template("{question}")    
        else: # 普通问答，带历史
            question=input
            prompt=PromptTemplate.from_template(prompt_history_template) if history is not None else PromptTemplate.from_template("{question}")
           
        chain=LLMChain(llm=run_llm, prompt=prompt,llm_kwargs={"temperature":temp})
        start=time.time()
        response=chain.run({"history":history,"question":question,"similarity":similarity})
        cost_time=time.time()-start
        print("cost:", round(cost_time, 2), "s")
        return response, [], input
        
    def chat(self, input, isExpert=False, isEnhancement=False, isBase=None, dialog=[], temp=0.8):
        if isExpert:
            print("专家指令",end=" ")
            if self.source_prefix is not None:
                history = []
                prompt = self.source_prefix + input
            else:
                prompt = input
            input = f"[专家指令]{input}"

        elif isEnhancement:
            print("知识增强",end=" ")
            history = []
            similarity = self.vectorstore_faiss._join_document(self.vectorstore_faiss.get_text_similarity_with_score(input))
            if similarity is not None :
                prompt=f"{similarity}\n请根据上述内容回答以下问题：\n{input}"         
            elif self.source_prefix is not None:
                prompt = self.source_prefix + input
            else:
                prompt = input
            input = f"[知识增强]{input}"

        else:
            print("普通问答",end=" ")
            if self.re_history is not None:
                history = self.re_history(dialog)
            else:
                history = dialog
            prompt = input
            input = f"[普通问答]{input}"

        if isBase is not None and isBase:
            print("基础模型")
            exec_model, exec_tokenizer = self.base_model, self.base_tokenizer
        else:
            print("增强模型")
            exec_model, exec_tokenizer = self.model, self.tokenizer

        start=time.time() 
        # response, history = exec_model.chat(
        #     exec_tokenizer, prompt, history, temperature=temp)

        prompt = self._build_prompt(prompt, history)
        response = exec_model(prompt)

        cost_time=time.time()-start
        
        if exec_tokenizer is not None:
            input_token_size = len(exec_tokenizer.encode(input))
            output_token_size = len(exec_tokenizer.encode(response))
            
            print()
            print("【Itoken_size:", input_token_size, "】prompt:", prompt)
            print("【Otoken_size:", output_token_size, "】response:", response)
            print("cost:", round(cost_time, 2), "s", 
                "tps:", round((input_token_size + output_token_size / cost_time), 2),"token/s")
            print("--------------------------------------------------")
        else:
            print("cost:", round(cost_time, 2), "s")
            print("--------------------------------------------------")
        
        return response, history, input