import os, sys
sys.path.append("../..") 

import json
import pandas
from llm.spark import SparkLLM
from llm.ernie import ErnieLLM, ModelType
from qa.generator import QAGenerator as QAGeneratorBase
from qa.generator import TrainData
from langchain.chat_models import ChatOpenAI
from langchain.base_language import BaseLanguageModel

class QAGenerator:
    def __init__(self, llm: BaseLanguageModel):
        """
        初始化 QAGenerator

        :param llm: 语言模型
        """
        self.generator = TrainData(llm=llm)

    def generate_questions_and_answers(self, input_text, question_number=3):
        """
        生成问题和答案

        :param input_text: 输入文本
        :param question_number: 生成问题的数量，默认为 3
        :return: 问题和答案的列表
        """
        # questions = self.generator.generate_questions(input_text, question_number=question_number)
        # result = []
        # for q in questions:
        #     answer = self.generator.generate_answer(input_text, q)
        #     result.append((q, answer))
        result = self.generator.generate(input_text, question_number=question_number)
        return result

    def read_and_deduplicate_text(self, data_path, text_column):
        """
        读取指定路径下的文件并返回去重后的文本列表

        :param data_path: 文件路径
        :param text_column: 文本所在列的列名
        :return: 去重后的文本列表
        """
        with open(data_path, "r", encoding="utf-8") as f:
            text_list=[]
            ext = data_path.split(".")[-1]
            if ext == "json":
                data = json.load(f)
                for item in data:
                    text_list.append(item[text_column])
            elif ext == "csv":
                data = pandas.read_csv(f)
                for item in data[text_column]:
                    text_list.append(item)

            # 去重
            text_list = list(set(text_list))
            print(f"共读取到 {len(text_list)} 条文本")
            return text_list

    def save_questions_and_answers(self, df, output_path):
        """
        将问题和答案保存到 CSV 文件中，已追加的方式

        :param df: 问题和答案的 DataFrame
        :param output_path: 输出文件路径
        """
        if os.path.exists(output_path):
            df.to_csv(output_path, mode="a", index=False, header=False)
        else:
            df.to_csv(output_path, index=False)


    def generate_questions_csv(self, data_path, text_column, output_path, save_step=100, max_num=1000):
        """
        生成问题和答案的 CSV 文件

        :param data_path: 数据文件路径
        :param text_column: 文本所在列的列名
        :param output_path: 输出文件路径
        """
        # 读取数据文件并去重
        text_list = self.read_and_deduplicate_text(data_path, text_column)
        text_list=text_list[:max_num]
        # 生成问题和答案
        result = []
        for i, input in enumerate(text_list):
            if len(input) > 200:
                num_questions = 5
            else:
                num_questions = 3
            questions = self.generate_questions_and_answers(input, question_number=num_questions)
            for q, a in questions:
                result.append((input, q, a))
            if (i+1) % save_step == 0:
                questions_df = pandas.DataFrame(result, columns=["text", "question", "answer"])
                self.save_questions_and_answers(questions_df, output_path)
                result = []

            # 打印进度条
            progress = (i + 1) / len(text_list) * 100
            bar_length = 50
            filled_length = int(bar_length * progress // 100)
            bar = "█" * filled_length + "-" * (bar_length - filled_length)
            print(f"\r已处理 {i+1}/{len(text_list)} 条数据 |{bar}| {progress:.2f}%", end="")


        # 将结果保存到 CSV 文件中
        questions_df = pandas.DataFrame(result, columns=["text", "question", "answer"])
        self.save_questions_and_answers(questions_df, output_path)
        # questions_df = pandas.DataFrame(result, columns=["text", "question", "answer"])
        # questions_df.to_csv(output_path, index=False)
